import netCDF4
import numpy
import re
import os
import copy
from .field import (Field, Space, FieldList, Coordinate, CoordinateBounds, 
                    Transform, CellMeasures, CellMethods, CoordinateList)
from .data  import Data, set_Data_attributes, dimension_direction
from .utils import CfDict

# Initialize global variables (see below for documentation)
seen_in_file = {}
increasing   = {}

# Global variable   Description
# ----------------- ------------------------------------------------------
# seen_in_file
# increasing      
# ----------------- ------------------------------------------------------

def read(filename, close=False, exact=False, extract={}):
    ''' 

    Read fields from an input netCDF file on disk or from an OPeNDAP
    server location.
    
    All data arrays from all field components are not read from disk
    but are stored in the returned fields as file pointers.

    NetCDF dimension names are stored in the :attr:`nc_dimensions`
    attribute of a field's space and netCDF variable names are stored
    in the :attr:`ncvar` attributes of the field and its space
    components (coordinates, coordinate bounds, cell measures and
    transformations).
    

    :argument filename: A string giving the file name or OPenDAP URL
                        from which to read fields.

    :type filename: `str`

 
    :returns: A list of fields (a `cf.FieldList` object).

    **Examples**
 
    >>> s = cf.netcdf.read('file.nc')
    >>> type(s)
    <class 'cf.field.FieldList'>
    >>> s
    [<CF Field: pmsl(30, 24)>,
     <CF Field: z-squared(17, 30, 24)>,
     <CF Field: temperature(17, 30, 24)>,
     <CF Field: temperature_wind(17, 29, 24)>]

    >>> cf.netcdf.read('file.nc')[0:2]
    [<CF Field: pmsl(30, 24)>,
     <CF Field: z-squared(17, 30, 24)>]

    >>> cf.netcdf.read('file.nc', units='K')
    [<CF Field: temperature(17, 30, 24)>,
     <CF Field: temperature_wind(17, 29, 24)>]

    >>> cf.netcdf.read('file.nc').extract(units='K')
    [<CF Field: temperature(17, 30, 24)>,
     <CF Field: temperature_wind(17, 29, 24)>]

    >>> cf.netcdf.read('file.nc')[0]
    <CF Field: pmsl(30, 24)>


    .. seealso:: `~cf.read` for more examples of field selection by
                 phenomena criteria.

    '''
    global seen_in_file, increasing, formula_terms, global_attr
    
    # Read the netCDF file
    nc = netCDF4.Dataset(filename, 'r')

#    filename = os.path.basename(filename) 

    # ----------------------------------------------------------------
    # Put the file's global attributes into the global 'global_attr'
    # dictionary
    # ----------------------------------------------------------------
    global_attr = {}
    for attr in map(str, nc.ncattrs()):
        try:
            value = nc.getncattr(attr)
            if isinstance(value, basestring):
                try:
                    global_attr[attr] = str(value)
                except UnicodeEncodeError:
                    global_attr[attr] = value           
            else:
                global_attr[attr] = value     
        except UnicodeDecodeError:
            pass
    #--- End: for
    
    # ----------------------------------------------------------------
    # Create a list of the names of all of the variables in the
    # file. E.g. variables=['lon','lat','tas']
    # ----------------------------------------------------------------
    variables = map(str, nc.variables)
    
    # ----------------------------------------------------------------
    # Create a dictionary keyed by nc variable names where each key's
    # value is a dictionary of that variable's nc
    # attributes. E.g. var_attr['tas']['units']='K'
    # ----------------------------------------------------------------
    var_attr = {}
    for ncvar in variables:
        var_attr[ncvar] = {}
        for attr in map(str, nc.variables[ncvar].ncattrs()):
            try:
                var_attr[ncvar][attr] = nc.variables[ncvar].getncattr(attr)
                if isinstance(var_attr[ncvar][attr], basestring):
                    try:
                        var_attr[ncvar][attr] = str(var_attr[ncvar][attr])
                    except UnicodeEncodeError:
                        var_attr[ncvar][attr] = var_attr[ncvar][attr]                   
            except UnicodeDecodeError:
                pass
    #--- End: for

    # ----------------------------------------------------------------
    # Remove everything bar data variables from the list of
    # variables. I.e. remove dimension and auxiliary coordinates,
    # their bounds and grid_mapping variables
    # ----------------------------------------------------------------
    for ncvar in variables[:]:

        # Remove dimension coordinates and their bounds
        if ncvar in map(str, nc.dimensions):

            try:
                variables.remove(ncvar)
            except ValueError:
                # This ncvar has already been removed. For example, it
                # could have been an auxiliary coordinate to another
                # variable
                pass
            else:
                for attr in ('bounds', 'climatology'):
                    try:
                        bounds = var_attr[ncvar][attr]
                    except KeyError:
                        continue
                
                    # Check the dimensionality of the coordinate's
                    # bounds. If it is not right, then it can't be a
                    # bounds variable and so promote to an independent
                    # data variable
                    try:
                        b = nc.variables[bounds]
                    except KeyError:                    
                        print "Missing bounds variable '%(bounds)s' in %(filename)s" % \
                            locals()
                        del var_attr[ncvar][attr]
                    else:
                        if b.ndim == nc.variables[ncvar].ndim+1:
                            variables.remove(bounds)
                        else:
                            del var_attr[ncvar][attr]
                        break
                #--- End: for
            #--- End: try
            continue
        #--- End: if
    
        # Remove auxiliary coordinates and their bounds
        try:
            var_attr[ncvar]['coordinates']
        except KeyError:
            pass
        else:
            # Allow for (incorrect) comma separated lists
            for aux in re.split('\s+|\s*,\s*', var_attr[ncvar]['coordinates']):
                try:
                    variables.remove(aux)
                except ValueError:
                    pass
                else:
                    for attr in ('bounds', 'climatology'):
                        try:
                            bounds = var_attr[aux][attr]
                        except KeyError:
                            continue

                        # Check the dimensionality of the coordinate's
                        # bounds. If it is not right, then it can't be
                        # a bounds variable and so promote to an
                        # independent data variable
                        try:
                            b = nc.variables[bounds]
                        except KeyError:                    
                            print "Missing bounds variable '%(bounds)s' in %(filename)s" % \
                                locals()
                            del var_attr[aux][attr]
                        else:
                            if b.ndim == nc.variables[aux].ndim+1:
                                variables.remove(bounds)
                            else:
                                del var_attr[aux][attr]
                            break
                    #--- End: for
            #--- End: for
        #--- End: try
    
        # Remove grid mapping variables
        try:
            variables.remove(var_attr[ncvar]['grid_mapping'])
        except (KeyError, ValueError):
            pass
        #--- End: try
    #--- End: for

    # ----------------------------------------------------------------
    # Everything left in the variables list is now a proper data
    # variable, so make a list of fields, each of which contains one
    # data variable and the relevant shared metadata. Reset the global
    # variable 'seen_in_file' to an empty dictionary.
    # ----------------------------------------------------------------
    seen_in_file            = {}
    increasing              = {}
    formula_terms           = {}
    all_fields_in_file      = FieldList()
    selected_fields_in_file = FieldList()
    for data_var in variables:

        s = _create_a_field(nc, data_var, var_attr)
        
        s.setpriv('file', filename)
        
        all_fields_in_file.append(s)

        if s._match(exact, **extract):
            selected_fields_in_file.append(s)                                    
    #--- End: for

    # ----------------------------------------------------------------
    # If any transform equation terms are other fields then find out
    # which fields they are
    # ----------------------------------------------------------------
    if formula_terms:

        #find the ft fields
        ddd    = {}
        ft_ncvars = []
        for transform in formula_terms.values():
            for term, value in transform.iteritems():
                try:
                    ft_ncvars.append(value['ncvar'])
                except (TypeError, KeyError):
                    continue

        index = 0
        while ft_ncvars:
            try:
                s = selected_fields_in_file[index]
            except IndexError:
                break

            # Still here?
            ncvar = s.getpriv('ncvar')            
            if ncvar in ft_ncvars:
                ft_ncvars.remove(ncvar)
                ddd[ncvar] = s
                selected_fields_in_file.pop(index)
            else:
                index += 1
        #--- End: while

        for s in selected_fields_in_file._list:
            try:
                transforms = s.space.transform.values()
            except AttributeError:
                continue
        
            # Still here? Then loop round this field's transforms
            for transform in transforms:
                try:
                    transform['grid_mapping_name']
                    continue
                except KeyError:
                    pass

                # Still here? Then loop round this formula_terms
                # transform's terms and values
                for term, value in transform.iteritems():
                    try:
                        transform[term] = ddd[value['ncvar']] #.copy()
                    except (TypeError, KeyError): #(IndexError, TypeError, KeyError):
                        continue
                #--- End: for
            #--- End: for
        #--- End: for
    #--- End: if
        
    # ----------------------------------------------------------------
    # Replace string-valued ancillary variables items with fields. If
    # no field can be found then leave the element as a string.
    # ----------------------------------------------------------------
    # Loop round fields in this file
    for s in selected_fields_in_file:
        try:
            s.getpub('ancillary_variables')
        except AttributeError:
            continue
    
        # Loop round this field's ancillary_variables key, value
        # pairs
        i = -1
        for value in s.getpub('ancillary_variables')[:]:
            i += 1
            # Loop round every field in this file to see if
            # one of them is the required ancillary variable
            for t in all_fields_in_file:
#                if t.getpriv('ncvar') == value:
                if t.ncvar == value:
                    # This deep copy is ok, as we've not read
                    # any of the data yet and so it will
                    # contain the file pointers.
                    s.getpub('ancillary_variables')[i] = t.copy()
                    break
    #--- End: for

#    print  selected_fields_in_fil
#    print selected_fields_in_file.privattrs
#    z = selected_fields_in_file.copy()
#    print selected_fields_in_file.match(units='^K$')
 #   print repr(selected_fields_in_file.extract(units='^K$'))
#    print repr(z)

#    print selected_fields_in_file.type()
   
      
    # Close the file
    if close:
        nc.close()
    
    return selected_fields_in_file
#--- End: def


def _create_a_field(nc, data_var, var_attr):
    '''Private function'''
    
    def _create_Coordinate(f, key, nc, ncvar, var_attr):
        '''

Create a coordinate variable
'''
        global increasing

        c = Coordinate(public_attributes  = var_attr[ncvar],
                       private_attributes = {'ncvar': ncvar}) 
        c.ncvar = ncvar
        
        data = nc.variables[ncvar]
 
        # ------------------------------------------------------------
        # Collapse a string valued auxiliary
        # coordinate. E.g. ['a','b','c'] becomes ['abc']
        # ------------------------------------------------------------
        if data.dtype.kind is 'S':
            new_shape = data.shape[0:-1]
            new_size  = numpy.prod(new_shape)
            strlen    = data.shape[-1]
            data      = data[...].reshape((new_size, strlen))
            
            # Make sure everything is ok if data is a masked array
            try:
                data.set_fill_value(' ')
                data = data.filled()
            except AttributeError:
                pass

            new = [''.join(x).rstrip() for x in data]
            new = numpy.array(new, dtype='|S%(strlen)d' % locals())

            data = new.reshape(new_shape)

            del new
        #--- End: if

        c._data = Data(data)

        # ---------------------------------------------------------
        # Add any bounds
        # ---------------------------------------------------------
        for attr in ('bounds', 'climatology'):
            
            try:
                ncbounds = c.getpub(attr)
            except AttributeError:
                continue

            b = CoordinateBounds(public_attributes  = var_attr[ncbounds],
                                 private_attributes = {'ncvar': ncbounds})
            b.ncvar = ncbounds
            try:
                b.delpub('formula_terms')
            except AttributeError:
                pass
                
            b._data = Data(nc.variables[ncbounds])

            c.bounds = b

            break
        #--- End: for
        
        try:
            c.getpub('climatology')
        except AttributeError:
            pass
        else:
            c.climatology = True

        for attr in ('bounds', 'climatology'):            
            try:
                c.delpub(attr)
            except AttributeError:
                pass
        #--- End: for

        if key.startswith('d'): # d for dim
            direction              = dimension_direction(c)
            increasing[ncvar]      = direction
            f.space.increasing[key] = direction
        
        # ---------------------------------------------------------
        # Set the coordinate and bounds' Data objects' metadata
        # ---------------------------------------------------------
        c = set_Data_attributes(f, key, c)
        
        # ---------------------------------------------------------
        # Return the coordinate
        # ---------------------------------------------------------
        return c
    #--- End: def

    def _create_CellMeasures(f, key, nc, ncvar, var_attr):
        '''

Create a cell measures variable
'''

        cell = CellMeasures(public_attributes  = var_attr[ncvar],
                            private_attributes = {'ncvar': ncvar})
        cell.ncvar = ncvar

        cell._data = Data(nc.variables[ncvar])
        
        # ---------------------------------------------------------
        # Set the cell measures' Data objects' metadata
        # ---------------------------------------------------------
        cell = set_Data_attributes(f, key, cell)
        
        return cell
    #--- End: def

    # --------------------------------------------------------
    # Global variables
    # --------------------------------------------------------
    global seen_in_file, increasing, formula_terms, global_attr

    # Add global attributes to the data variable's properties, unless
    # the data variables already has a property with the same name.
    for attr, value in global_attr.iteritems():
        try:
            var_attr[data_var][attr]
        except KeyError:
            var_attr[data_var][attr] = value

    # ----------------------------------------------------------------
    # Initialize the field with the data variable and its attributes
    # ----------------------------------------------------------------
    t = Field(public_attributes  = var_attr[data_var],
              private_attributes = {'ncvar': data_var})
    t.ncvar = data_var

    t.space = Space()
    t._data = Data(nc.variables[data_var])

    # Initialize mappings of nc variable and dimension names to
    # internal field names. E.g. ncvar_to_var['geo_region']='aux0',
    # ncdim_to_dim['depth']='dim1'
    ncvar_to_var       = {}
    ncdim_to_dim       = {}
    t.space.increasing = {}
        
    t.space.dimensions['data'] = []
    t.space.nc_dimensions      = {}
    
    # ----------------------------------------------------------------
    # Add dimensions and dimension coordinates to the field
    # ----------------------------------------------------------------
    dimN = 0
    uncN = 0

    for ncdim in map(str, nc.variables[data_var].dimensions):
        
        # Set the internal field name for this dimension and map the
        # nc name to it
        dim = 'dim%(dimN)d' % locals()
        dimN += 1         # Increment dimension number
        ncvar_to_var[ncdim] = dim
        ncdim_to_dim[ncdim] = dim
        
        # Update data dimension name and set dimension size
        t.space.nc_dimensions[dim]   = ncdim
        t.space.dimension_sizes[dim] = len(nc.dimensions[ncdim])
        t.space.dimensions['data'].append(dim)
        
        # If a dimension coordinate exists then create it in the field
        if nc.variables.has_key(ncdim):

            t.space.dimensions[dim] = [dim]

            try:
                t.space[dim] = copy.deepcopy(seen_in_file[ncdim])
                t.space.increasing[dim] = increasing[ncdim]
            except KeyError: 
                pass
                coord = _create_Coordinate(t, dim, nc, ncdim, var_attr)

                seen_in_file[ncdim] = coord                
                t.space[dim]        = coord
        
            # --------------------------------------------------------
            # Compression by gathering
            # --------------------------------------------------------
            try:
                t.space[dim].compress = CoordinateList(t.space[dim].compress.split())
            except AttributeError:
                pass
            else:
                # Create the uncompressed coordinates
                for i in xrange(len(t.space[dim].compress)):
                    try:
                        t.space[dim].compress[i] = copy.deepcopy(seen_in_file[ncvar])
                    except KeyError:
                        pass
                        t.space[dim].compress[i] = _create_Coordinate(t, dim, 
                                                                     nc, ncvar, var_attr)
                        seen_in_file[ncvar] = t.space[dim].compress[i]
                #--- End: for

#                t.space[dim].compress = t.space[dim].compress.split()
#                for i in xrange(len(t.space[dim].compress)):
#                    # Create the coordinate without updating the
#                    # dimension_sizes or dimensions dictionaries
#                    unc = 'unc'+str(uncN)
#                    ncvar = t.space[dim].compress[i]
#                    t.space[dim].compress[i] = unc
#                    t.space.nc_dimensions[unc] = ncvar
#                    try:
 #                        t.space[unc] = copy.deepcopy(seen_in_file[ncvar])
#                    except KeyError:
#                        t.space[unc] = _create_Coordinate(nc, ncvar, var_attr)
#                        seen_in_file[ncvar] = t.space[unc] 
#
#                    # Increment ?? number
#                    uncN += 1
#                #--- End: for

        else:
            # Dimension coordinate for this dimension does not exist
            increasing[ncdim]       = True
            t.space.increasing[dim] = True
        #--- End: if

    #--- End: for

    # ----------------------------------------------------------------
    # Add auxiliary coordinates to the field
    # ----------------------------------------------------------------
    auxN = 0
    try:
        coordinates = t.getpub('coordinates')
    except AttributeError:
        pass
    else:
        # Allow for (incorrect) comma separated lists
        for ncvar in re.split('\s+|\s*,\s*', coordinates):

            # Skip dimension coordinates which are in the list
            if ncvar in nc.variables[data_var].dimensions:
                continue

            # Set the internal field name for this auxiliary
            # coordinate and map the nc name to it
            aux = 'aux%(auxN)d' % locals()
            auxN += 1       # Increment auxiliary coordinate number
            ncvar_to_var[ncvar] = aux

            # If the auxiliary coordinate variable exists then create
            # a new coordinate variable for it in the field dictionary
            if nc.variables.has_key(ncvar):

                # Set auxiliary coordinate dimension names
                t.space.dimensions[aux] = []
                for ncdim in nc.variables[ncvar].dimensions:
                    try:
                        t.space.dimensions[aux].append(ncdim_to_dim[ncdim])
                    except KeyError:
                        # Why try here? because of things like bounds, strlen
                        pass
                #--- End: for

                key = aux

                # ----------------------------------------------------
                # Create a dimension for a numeric scalar coordinate
                # and turn it inot a dimension coordinate.
                # ----------------------------------------------------
                if not t.space.dimensions[aux]:
                    dim = 'dim%(dimN)d' % locals()
                    dimN += 1         # Increment dimension number
                    
                    t.space.dimension_sizes[dim] = 1
                    t.space.dimensions[dim]      = [dim]

                    if nc.variables[ncvar].dtype.kind is 'S':
                        # String valued scalar coordinate                        
                        t.space.dimensions['aux'] = dim
                        t.space.increasing['dim'] = True
                    else:  
                        # Numeric valued scalar coordinate                           
                        del t.space.dimensions[aux]
                        key = dim
                #--- End: if

                try:                   
                    t.space[key] = copy.deepcopy(seen_in_file[ncvar])
                except KeyError:
                    coord = _create_Coordinate(t, key, nc, ncvar, var_attr)

                    seen_in_file[ncvar] = coord
                    t.space[key]         = coord
                #--- End: try
            #--- End: if

        #--- End: for

        t.delpub('coordinates')
    #--- End: try

    # ----------------------------------------------------------------
    # Now that we've created all of the dimension and auxiliary
    # coordiantes, set the field's Data objects's metadata
    # ----------------------------------------------------------------
    t = set_Data_attributes(t, 'data', t)

    # -----------------------------
    # Add cell methods to the field
    # -----------------------------
    try:
        # Try to replace the cell_methods attribute with a CellMethods
        # instance
        cell_methods   = CellMethods()
        t.cell_methods = cell_methods.parse(t.cell_methods, field=t)
    except AttributeError:
        # The field has no cell methods
        pass

    # ----------------------------------------------------------------
    # Add formula_terms to coordinates' Transform attributes
    # ----------------------------------------------------------------
    transN = 0
    for coord in t.space.keys():
 
#        coord_ncvar = t.space[coord].getpriv('ncvar')
        coord_ncvar = t.space[coord].ncvar

        try:
            transform = formula_terms[coord_ncvar]
        except KeyError:
            pass
        else:
            # This coordinate's formula_terms transform has already
            # been created, so just copy it            
            try:
                transform_id
            except UnboundLocalError:
                transform_id = 'trans%(transN)d' % locals()
                transN += 1       # Increment auxiliary coordinate number
            t.space[coord].transform = transform_id
            try:
                t.space.transform
            except AttributeError:
                t.space.transform = CfDict()
            t.space.transform[transform_id] = copy.deepcopy(transform)
            continue
        #-- End: try

        # Still here?
        try:
            # This coordinate has a CF formula_terms attribute
            form_terms = t.space[coord].getpub('formula_terms')
        except AttributeError:
            # This coordinate doesn't have a formula_terms attribute
            continue

        # ------------------------------------------------------------
        # Still here? 
        # ------------------------------------------------------------
        transform      = Transform()
        transform.name = t.space[coord].standard_name
        
        # Add the equation terms and references to their values to to
        # new auxiliary coordinate's transform.
        ft = re.split('\s+|\s*:\s+', form_terms)
        for term, ncvar in zip(ft[0::2], ft[1::2]):
            try:
                # Try a coordinate variable from this field
                value = ncvar_to_var[ncvar]
            except KeyError:
                # Assume it's another field. We don't know what this
                # field is yet, so we'll give it its ncvar name and
                # work out which field it is later, if we
                # can. E.g. value={'ncvar':'pstar'}
                value = {'ncvar': ncvar}
    
            transform[term] = value
        #--- End: for 

        transform_id = 'trans%(transN)d' % locals()
        transN += 1       # Increment transform number

        t.space[coord].transform = transform_id

        try:
            t.space.transform
        except AttributeError:
            t.space.transform = CfDict()

        t.space.transform[transform_id] = transform

        formula_terms[coord_ncvar] = transform

        t.space[coord].delpub('formula_terms')
    #--- End: for

    # ----------------------------------------------------------------
    # Add grid mapping to coordinates' Transform attributes
    # ----------------------------------------------------------------
    try:
        grid_mapping = t.getpub('grid_mapping')
    except AttributeError:
        # No grid mapping for this field
        pass
    else:
        # Got a grid_mapping for this field
        try:
            t.space.transform
        except AttributeError:
            t.space.transform = CfDict()                

        for gm in grid_mapping.split():

            try:
                var_attr[gm]
            except KeyError:
                continue

            try:
                var_attr[gm]['grid_mapping_name']
            except KeyError:
                var_attr[gm]['grid_mapping_name'] = gm
            grid_mapping_name = var_attr[gm]['grid_mapping_name']

            transform_id = 'trans%(transN)d' % locals()
            transN += 1     

            t.space.transform[transform_id] = Transform(**var_attr[gm])
            t.space.transform[transform_id].ncvar = gm
            t.space.transform[transform_id].name  = grid_mapping_name

            # Work out which coordinates need the transform attached
            # to them
            if grid_mapping_name in ('rotated_latitude_longitude'):
                for coord in t.space.get_keys('^dim|^aux'):
                    if t.space[coord].name() in ('longitude', 
                                                'latitude',
                                                'grid_longitude',
                                                'grid_latitude'):
                        t.space[coord].transform = transform_id

            elif grid_mapping_name == 'latitude_longitude':
                for coord in t.space.get_keys('^dim|^aux'):
                    if t.space[coord].name() in ('longitude', 
                                                'latitude'):
                        t.space[coord].transform = transform_id

            else:
                for coord in t.space.get_keys('^dim|^aux'):
                    if t.space[coord].name() in ('projection_x_coordinate', 
                                                'projection_y_coordinate'):
                        t.space[coord].transform = transform_id
            #--- End: if
        #--- End: for

        t.delpub('grid_mapping')
    #--- End: try

    # ----------------------------------------------------------------
    # Add cell measures to the field
    # ----------------------------------------------------------------
    try:
        # Try to parse a cell measures attribute
        cell_measures = re.split('\s*(\w+):\s*', t.getpub('cell_measures'))
    except AttributeError:
        # No cell measures for this field
        pass
    else:
        # Got cell measures
        cell_measures.pop(0)
        cmN = 0
        for measure, ncvar in zip(cell_measures[::2], 
                                  cell_measures[1::2]):

            cm = 'cm%(cmN)d' % locals()
            cmN += 1                # Increment cell measures number

            # Set cell measures' dimension names
            try:
                t.space.dimensions[cm] = [ncdim_to_dim[ncdim]
                                         for ncdim in nc.variables[ncvar].dimensions]
            except KeyError:
                # This cell measure variable does exist in the file
                cmN -= 1
                continue

            try:
                t.space[cm] = copy.deepcopy(seen_in_file[ncvar])
            except KeyError:

                try:             
                    var_attr[ncvar]
                except KeyError:
                    del t.space.dimensions[cm]
                else:                        
                    cell = _create_CellMeasures(t, cm, nc, ncvar, var_attr)
                    
                    try:
                        cell.standard_name
                    except AttributeError:
                        cell.standard_name = measure

                    # ------------------------------------------------
                    # Keep a record of field components which exist in
                    # this file so that they may be shared easily
                    # ------------------------------------------------
                    seen_in_file[ncvar] = cell
                    t.space[cm]         = cell
            #--- End: try    
        #--- End: for

        t.delpub('cell_measures')
    #--- End: try

    # Make dimension coordinate with no aux, cell_measures scalar and
    # drop their dimensions from the data
    t.squash()

    # ----------------------------------------------------------------
    # Parse ancillary variables to a FieldList of ncvar names, which
    # will get converted to fields later # dch why not a list?
    # ----------------------------------------------------------------
    try:
        t.setpub('ancillary_variables',
                 FieldList(t.getpub('ancillary_variables').split()))
    except AttributeError:
        pass

    # Convert flag_meanings to a list
    try:
        t.setpub('flag_meanings', t.getpub('flag_meanings').split())
    except AttributeError:
        pass

    del t.space.increasing

    # Return the finished field
    return t
#--- End: def
