import os
import numpy
import netCDF4
import csv
import re
import textwrap
import time
from operator import mul

from .field import (Field, FieldList, Space, CoordinateBounds, 
                    Coordinate, CellMethods, Transform)
from .utils import CfDict, RTOL, ATOL, array_allclose, equals
from .data  import Data, set_Data_attributes, dimension_direction
from . import __version__, __Conventions__, __file__

_word_size = 4
_pp_rmdi   = -1e+30 

# --------------------------------------------------------------------
# Set default UM version from the environment
# --------------------------------------------------------------------
try:
    _UMVERSION = os.environ['UMVERSION']
    _UMVERSION = _UMVERSION.replace('.', '0', 1)
except KeyError:
    pass

# --------------------------------------------------------------------
# Constants for converting creating unrotated lat/lon auxiliary
# coordinates
# --------------------------------------------------------------------
_PI          = numpy.pi
_PI_OVER_180 = _PI/180.0

# --------------------------------------------------------------------
# Number matching regular expression
# --------------------------------------------------------------------
_number_regex = '([-+]?\d*\.?\d+(e[-+]?\d+)?)'

# --------------------------------------------------------------------
# Datetime object that copes with non-standard calendars
# --------------------------------------------------------------------
datetime = netCDF4.netcdftime.datetime

# --------------------------------------------------------------------
# Cache for derived values
# --------------------------------------------------------------------
_cache = {'runid' : {},
          'latlon': {},
          'time'  : {},
          }

# --------------------------------------------------------------------
# Names of PP integer and real header items
# --------------------------------------------------------------------
_lhd_items = ('lbyr', 'lbmon', 'lbdat', 'lbhr', 'lbmin', 'lbday',
              'lbyrd', 'lbmond', 'lbdatd', 'lbhrd', 'lbmind',
              'lbdayd', 'lbtim', 'lbft', 'lblrec', 'lbcode', 'lbhem',
              'lbrow', 'lbnpt', 'lbext', 'lbpack', 'lbrel', 'lbfc',
              'lbcfc', 'lbproc', 'lbvc', 'lbrvc', 'lbexp', 'lbegin', 
              'lbnrec', 'lbproj', 'lbtyp', 'lblev', 'lbrsvd1',
              'lbrsvd2', 'lbrsvd3', 'lbrsvd4', 'lbsrce', 'lbuser1',
              'lbuser2', 'lbuser3', 'lbuser4', 'lbuser5', 'lbuser6',
              'lbuser7')

_bhd_items = ('brsvd1', 'brsvd2', 'brsvd3', 'brsvd4', 
              'bdatum', 'bacc', 'blev', 'brlev', 'bhlev', 'bhrlev',
              'bplat', 'bplon', 'bgor',
              'bzy', 'bdy', 'bzx', 'bdx', 'bmdi', 'bmks')

# --------------------------------------------------------------------
# Positions of PP header items in their arrays
# --------------------------------------------------------------------
(lbyr, lbmon, lbdat, lbhr, lbmin, lbday,
 lbyrd, lbmond, lbdatd, lbhrd, lbmind,
 lbdayd, lbtim, lbft, lblrec, lbcode, lbhem,
 lbrow, lbnpt, lbext, lbpack, lbrel, lbfc,
 lbcfc, lbproc, lbvc, lbrvc, lbexp, lbegin, 
 lbnrec, lbproj, lbtyp, lblev, lbrsvd1,
 lbrsvd2, lbrsvd3, lbrsvd4, lbsrce, lbuser1,
 lbuser2, lbuser3, lbuser4, lbuser5, lbuser6,
 lbuser7) = range(45)

(brsvd1, brsvd2, brsvd3, brsvd4, 
 bdatum, bacc, blev, brlev, bhlev, bhrlev,
 bplat, bplon, bgor,
 bzy, bdy, bzx, bdx, bmdi, bmks) = range(19)

# --------------------------------------------------------------------
# Tranlate PP axis codes to CF 'standard_name' values. (The full list
# of field code keys may be found at
# http://cms.ncas.ac.uk/html_umdocs/wave/@header.)
# --------------------------------------------------------------------
_coord_standard_name = {
    1   : 'air_pressure', 
    2   : 'altitude',
    3   : 'atmosphere_hybrid_sigma_pressure_coordinate',
    4   : 'depth',
    5   : 'model_level_number',        
    6   : 'air_potential_temperature',
    7   : 'atmosphere_sigma_coordinate',
    10  : 'latitude',
    11  : 'longitude',
    13  : 'region',
    14  : 'atmosphere_hybrid_height_coordinate',
    15  : 'height',
    20  : 'time',      # time (gregorian)
    23  : 'time',      # time (360_day)
    40  : 'pseudolevel',        # pseudolevel. THIS IS NOT A STANDARD NAME!!
    }

# --------------------------------------------------------------------
# Tranlate PP axis codes to CF 'long_name' values
# --------------------------------------------------------------------
_coord_long_name = {
    40 : 'pseudolevel',
    }

# --------------------------------------------------------------------
# Tranlate PP axis codes to CF 'units' values
# --------------------------------------------------------------------
_coord_units = {
    1   : 'hPa',           # air_pressure                      
    2   : 'm',             # altitude         
    3   : '1',             # atmosphere_hybrid_sigma_pressure_coordinate
    4   : 'm',             # depth                                  
    5   : '1',             # model_level_number                         
    6   : 'K',             # air_potential_temperature
    7   : '1',             # atmosphere_sigma_coordinate               
    10  : 'degrees_north', # latitude                               
    11  : 'degrees_east',  # longitude                                
    13  : 'string',        # region                                     
    14  : '1',             # atmosphere_hybrid_height_coordinate          
    15  : 'm',             # height                                      
    20  : 'days',          # time (gregorian)                    
    23  : 'days',          # time (360_day)
    40  : '1',             # pseudolevel
    }

# --------------------------------------------------------------------
# Tranlate PP axis codes to CF 'axis' values
# --------------------------------------------------------------------
_coord_axis = {
    1   : 'Z',   # air_pressure                       
    2   : 'Z',   # altitude                                     
    3   : 'Z',   # atmosphere_hybrid_sigma_pressure_coordinate  
    4   : 'Z',   # depth                                        
    5   : 'Z',   # model_level_number                          
    6   : 'Z',   # air_potential_temperature
    7   : 'Z',   # atmosphere_sigma_coordinate                
    10  : 'Y',   # latitude                                     
    11  : 'X',   # longitude                                    
    13  : None,  # region                                       
    14  : 'Z',   # atmosphere_hybrid_height_coordinate          
    15  : 'Z',   # height                                       
    20  : 'T',   # time (gregorian)                                         
    23  : 'T',   # time (360_day)                                         
    40  : None,  # pseudolevel                                    
    }

# --------------------------------------------------------------------
# Tranlate axis codes to CF 'positive' alues
# --------------------------------------------------------------------
_coord_positive = {
    1   : 'down',  # air_pressure                     
    2   : 'up',    # altitude                                  
    3   : 'down',  # atmosphere_hybrid_sigma_pressure_coordinate 
    4   : 'down',  # depth                                     
    5   : None,    # model_level_number                         
    6   : 'up',    # air_potential_temperature
    7   : 'down',  # atmosphere_sigma_coordinate               
    10  : None,    # latitude                                   
    11  : None,    # longitude                                   
    13  : None,    # region                                     
    14  : 'up',    # atmosphere_hybrid_height_coordinate         
    15  : 'up',    # height                                      
    20  : None,    # time (gregorian)                                          
    23  : None,    # time (360_day)                                        
    40  : None,    # pseudolevel                                    
    }

# --------------------------------------------------------------------
# Translate LBVC codes to PP axis codes. (The full list of field code
# keys may be found at
# http://cms.ncas.ac.uk/html_umdocs/wave/@fcodes.)
# --------------------------------------------------------------------
_lbvc_codes = {
    1   :  2,   # altitude (Height) 
    2   :  4,   # depth (Depth)
    3   : None, # (Geopotential (= g*height))
    4   : None, # (ICAO height)
    6   :  5,   # model_level_number  
    7   : None, # (Exner pressure)
    8   :  1,   # air_pressure  (Pressure)
    9   :  3,   # atmosphere_hybrid_sigma_pressure_coordinate (Hybrid pressure)
    10  :  7,   # atmosphere_sigma_coordinate (Sigma (= p/surface p))   ## dch check
    16  : None, # (Temperature T)
    19  :  6,   # air_potential_temperature (Potential temperature)
    27  : None, # (Atmospheric) density
    28  : None, # (d(p*)/dt .  p* = surface pressure)
    44  : None, # (Time in seconds)
    65  : 14,   # atmosphere_hybrid_height_coordinate (Hybrid height)
    176 : 10,   # latitude    (Latitude)
    177 : 11,   # longitude   (Longitude)
    }

# --------------------------------------------------------------------
# Characters used in decoding LBEXP into a runid
# --------------------------------------------------------------------
_characters = ('a','b','c','d','e','f','g','h','i','j','k','l','m',
               'n','o','p','q','r','s','t','u','v','w','x','y','z',               
               '0','1','2','3','4','5','6','7','8','9'
               )   

# --------------------------------------------------------------------
# Names of PP extra data codes
# --------------------------------------------------------------------
_extra_data_name = {
    1  : 'x',
    2  : 'y',
    3  : 'y_domain_lower_bound',
    4  : 'x_domain_lower_bound',
    5  : 'y_domain_upper_bound',
    6  : 'x_domain_upper_bound',
    7  : 'z_domain_lower_bound',
    8  : 'x_domain_upper_bound',
    10 : 'title',
    11 : 'domain_title',
    12 : 'x_lower_bound',
    13 : 'x_upper_bound',
    14 : 'y_lower_bound',
    15 : 'y_upper_bound',
    }

class Variable(object):
    ''' 
    
'''

    __slots__ = ['_file',
                 '_start',
                 '_shape',
                 '_dtype',
                 '_FillValue',
                 'scale_factor',
                 'add_offset',
                 ]


    def __init__(self, **kwargs):
        '''
'''
        for attr, value in kwargs.iteritems():
            setattr(self, attr, value) 
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        array = self.numpy_array
        return array[index]
    #--- End: def

    @property
    def shape(self):
        '''

Data's dimension sizes.

:returns: A tuple

    '''
        return self._shape
    #--- End: def

    @property
    def size(self):
        '''

Number of elements in the data.

:returns: A non-negative integer
    '''
        return reduce(mul, self._shape, 1) 
    #--- End: def

    @property
    def ndim(self):
        '''

Number of data dimensions.

:returns: A non-negative integer
    '''
        return len(self._shape)
    #--- End: def

    @property
    def dtype(self):
        '''

Data-type of the data elements.

:returns: A numpy dtype object.

    '''
        return self._dtype
    #--- End: def

    def close(self):
        '''
'''
        self._file.close()
    #--- End: def

    @property
    def numpy_array(self):
        '''

Read data from disk into a numpy array
'''
        # Make sure that the file is open
        closed = 0
        if self._file.closed:
            self._file = open(self._file.name, 'rb')
            closed = 1

        # Jump to the correct place in the file
        self._file.seek(self._start, os.SEEK_SET)

        # Read the data array
        array = numpy.fromfile(self._file,
                               dtype=self.dtype,
                               count=self.size).reshape(self.shape)

        # Re-close the file, if we had to open it
        if closed:
            self._file.close()

        # Convert to a masked array, if _FillValue is set and the
        # array contains missing data
        try:
            _FillValue = self._FillValue
        except AttributeError:
            pass
        else:
            array = numpy.ma.array(array, 
                                   mask=(array == _FillValue),
                                   fill_value=_FillValue)

        # Unpack the array using scale_factor and add_offset, if
        # available
        try:
            if self.scale_factor != 1.0:
                array *= self.scale_factor
        except AttributeError:
            pass
        
        try:
            if self.add_offset != 0.0:
                array += self.add_offset
        except AttributeError:
            pass
        
        # Return the array
        return array
    #--- End: def

    def equals(self, other, rtol=None, atol=None):

        # Compare variables which are the same
        if id(self) == id(other):
            return True
    
        # Compare object which are both of the same type
        if type(self) == type(other):

            # Check that each instance has the same relevant property
            # values
            if (self._file != other._file and not
                os.path.samefile(self._file.name, other._file.name)):
                return False

            for prop in ('_start', '_dtype', '_shape'):
                x = getattr(self, prop)
                y = getattr(other, prop)
                if x != y:
                    return False                

            # Still here? Then they're the same
            return True
    
        if isinstance(other, numpy.ndarray):
            if rtol is None:
                rtol = RTOL()
            if atol is None:
                atol = ATOL()   

            return array_allclose(self.numpy_array, other, rtol=rtol, atol=atol)
        #--- End: if
    #--- End: def

#--- End: class


class VariableCalc(Variable):
    '''

A PP variable which is specified by a start value and a constant
increment, such as a regular coordinate.
'''
    __slots__ = ['_pseudo_origin',
                 '_delta',
                 ]

    @property
    def numpy_array(self):
        '''

        :return: a numpy array '''
        array = self._pseudo_origin + \
            numpy.arange(1,self.size+1, dtype=self.dtype)*self._delta

        return array
    #--- End: def


    def equals(self, other, rtol=None, atol=None):

        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()

        if isinstance(other, numpy.ndarray):
            return array_allclose(self.numpy_array, other, rtol=rtol, atol=atol)

        # Check that each instance has the same relevant property
        # values
        for prop in ('_pseudo_origin', '_delta'):
            try:
                x = getattr(self, prop)
                y = getattr(other, prop)
                if not abs(x-y) <= atol + rtol*abs(y):
                    return False                
            except AttributeError:
                return False        

        try:
            x = self._shape
            y = other._shape
            if x != y:
                return False                
        except AttributeError:
            return False
        
        # Still here? Then they're the same
        return True

#        # Compare variables which are the same
#        if id(self) == id(other):
#            return True
    #--- End: def

#--- End: class


class VariableCalcBounds(VariableCalc):
    '''
'''
    __slots__ = []

    @property
    def numpy_array(self):

        # return coordinate bound values
        array      = numpy.empty(self.shape, dtype=self.dtype)

        array[:,0] = (self._pseudo_origin + 0.5*self._delta + 
                      numpy.arange(self.shape[0], dtype=self.dtype)*self._delta)

        array[:,1] = array[:,0] + self._delta
    
        return array
    #--- End: def

#--- End: class


class VariableBounds(Variable):
    '''  
'''
    __slots__ = ['_lower',
                 '_upper',
                 ]

    def __init__(self, lower, upper):
        '''
'''
        self._lower = lower
        self._upper = upper

        self._dtype = self._lower.dtype
        self._shape = self._lower.shape + (2,)        
    #--- End: def
   
    @property     
    def numpy_array(self):
        '''
'''
        try:
            lower = self._lower.numpy_array
        except AttributeError:
            lower = self._lower

        try:
            upper = self._upper.numpy_array
        except AttributeError:
            upper = self._upper

        return numpy.column_stack((lower, upper))
    #--- End: def
#--- End: class


class PPMetadata(object):

    def __init__(self):
        '''

Set attributes giving the default data type for reals and
integers. The basic assumption is that data are native endian.
'''
        self.float32      = numpy.dtype('<f4').newbyteorder('=')
        self.int32        = numpy.dtype('<i4').newbyteorder('=')
        self.str32        = numpy.dtype('<S4').newbyteorder('=')
        self.file_float32 = self.float32
        self.file_int32   = self.int32
        self.file_str32   = self.str32
    #--- End: def

    @property
    def printfdr(self):
        '''
        '''
        fdr = []
        for i in range(45):
            fdr.append('%s:_%s' % (_lhd_items[i].upper(), self.lhd[i]))
        for i in range(19):
            fdr.append('%s:_%s' % (_bhd_items[i].upper(), self.bhd[i]))
        fdr = textwrap.fill( ' '.join(fdr), width=79)
        fdr = [fdr.replace(':_', ': ')]

        if self.extra:
            fdr.append('EXTRA DATA:')
            for key in self.extra:
                fdr.append('%s: %s' % (key, repr(self.extra[key])))
        
        print '\n'.join(fdr)
    #--- End: def

#--- End: class


def unrotated_latlon(rotated_lat, rotated_lon, pole_lat, pole_lon):
    '''

Create 2-d arrays of unrotated latitudes and longitudes.
'''
    # Make sure rotated_lon and pole_lon is in [0, 360)
    pole_lon = pole_lon % 360.0

    # Cconvert everthing to radians
    pole_lon *= _PI_OVER_180
    pole_lat *= _PI_OVER_180

    cos_pole_lat = numpy.cos(pole_lat)
    sin_pole_lat = numpy.sin(pole_lat)

    # Create appropiate copies of the input rotated arrays
    rot_lon = rotated_lon.copy()
    rot_lat = rotated_lat.view()

    # Make sure rotated longitudes are between -180 and 180
    rot_lon %= 360.0
    rot_lon = numpy.where(rot_lon < 180.0,
                          rot_lon, rot_lon-360)

    # Create 2-d arrays of rotated latitudes and longitudes in radians
    nlat = rot_lat.size
    nlon = rot_lon.size
    rot_lon = numpy.resize(numpy.deg2rad(rot_lon), (nlat, nlon))    
    rot_lat = numpy.resize(numpy.deg2rad(rot_lat), (nlon, nlat))
    rot_lat = numpy.transpose(rot_lat, axes=(1,0))

    # Find unrotated latitudes
    CPART = numpy.cos(rot_lon) * numpy.cos(rot_lat)
    sin_rot_lat = numpy.sin(rot_lat)
    x = cos_pole_lat * CPART + sin_pole_lat * sin_rot_lat
    x = numpy.clip(x, -1.0, 1.0)
    unrotated_lat = numpy.arcsin(x)
    
    # Find unrotated longitudes
    x = -cos_pole_lat*sin_rot_lat + sin_pole_lat*CPART
    x /= numpy.cos(unrotated_lat)   # dch /0 or overflow here? surely
                                    # lat could be ~+-pi/2? if so,
                                    # does x ~ cos(lat)?
    x = numpy.clip(x, -1.0, 1.0)
    unrotated_lon = -numpy.arccos(x)
    
    unrotated_lon = numpy.where(rot_lon > 0.0, 
                                -unrotated_lon, unrotated_lon)
    if pole_lon >= ATOL():
        SOCK = pole_lon - _PI
    else:
        SOCK = 0
    unrotated_lon += SOCK

    # Convert unrotated latitudes and longitudes to degrees
    unrotated_lat = numpy.rad2deg(unrotated_lat)
    unrotated_lon = numpy.rad2deg(unrotated_lon)

    return unrotated_lat, unrotated_lon

    # Return unrotated latitudes and longitudes
#    return numpy.rollaxis(numpy.dstack((unrotated_lat, unrotated_lon)), 2)
#--- End: def

def _extra_data(ppfile, m=None):
    '''

Read the extra data (if any) of the PP field at the current position
in the given PP file, decode it and store it in the field's metadata
object.

:parameter m: A PP field's metatdata object
:type m: PPMetadata

Excerpt from UMDP F3 vn7.8
--------------------------
The data record may include 'extra data' in addition to the usual
grid-point values. Cross-section fields must always be followed by
extra data giving the x- and y-coordinate values of each column and
row of the grid. (This is to allow irregular grid). It is also
possible to use extra data in conjunction with other types of field,
eg. a character string description of the field. The header variable
LBEXT shows the length of extra data associated with a field. Thus the
total length (in words) of the data record (LBLREC) equals the number
of grid-point values plus LBEXT. Extra data is arranged in vectors,
each of which is made up of an integer code followed by data
values. The integer code is (1000xIA + IB), where IA is the length of
the vector in words (i.e. usually the number of data values) and IB is
a standard code. A zero integer code indicates no (more) extra data.
For cross-sections the extra data MUST start with an x-vector followed
by a y-vector; optionally other extra data may follow.
'''
    if m.lhd[lbext] <= 0:
        # Skip the 4th block control word
        ppfile.seek(_word_size, os.SEEK_CUR) 
        m.extra = {}
        return m

    extra = {}

    domain_titles = []

    len = 0        
    while len < m.lhd[lbext]:

        intcode = numpy.fromfile(ppfile, dtype=m.file_int32, count=1)

        # Stop if we have run out of extra data
        if not (intcode or intcode[0]):
            break

        # Find ia = size of this extra data variable, 
        #      ib = type of this extra data variable
        ia, ib = divmod(intcode[0], 1000)
        
        if ib in (10, 11):
            # Read characters from the file
            title = numpy.fromfile(ppfile, dtype=m.file_str32, count=ia)
            title = ''.join(title)
            if ib == 10:
                extra[_extra_data_name[ib]] = numpy.array([title])
            else:
                domain_titles.append(title)
        else:
            # Don't read an array of number, but store it as file
            # pointer
            parameters = {'_file'     : ppfile,
                          '_start'    : ppfile.tell(),
                          '_dtype'    : m.file_float32,
                          '_shape'    : (ia,),
                          }        
            extra[_extra_data_name[ib]] = Variable(**parameters)
            
            # Skip to the next extra data array
            ppfile.seek(ia*_word_size, os.SEEK_CUR)
        #--- End: if

        len += ia + 1
    #--- End: while    

    for bounds_type in ('', '_domain'):

        for dim in ('x', 'y'):
            dim_bounds_type = '%s%s' % (dim, bounds_type)
            lower  = '%s_lower_bound' % dim_bounds_type
            upper  = '%s_upper_bound' % dim_bounds_type
            bounds = '%s_bounds'      % dim_bounds_type
            
            try:
                extra[lower]
                extra[upper]
            except KeyError:
                pass
            else:
                extra[bounds] = VariableBounds(extra[lower], extra[upper])         
                extra.pop(lower)
                extra.pop(upper)
    #--- End: for

    # If we have domain titles then add them to the extra data
    # dictionary as a numpy array
    if domain_titles:
        extra[_extra_data_name[11]] = numpy.array(domain_titles)

    # Skip the 4th block control word so that we're ready to read the
    # next PP field
    ppfile.seek(_word_size, os.SEEK_CUR)            

    # Add the extra data to the metadata object
    m.extra = extra

    # Return the updated metadata object
    return m
#--- End: def

def _field_data(ppfile, m=None):    
    '''

Return a file pointer to the field's 2-d data array.

:parameter ppfile: dch 

:type ppfile: file

:parameter m: A PP field's complete metatdata
:type m: PPMetadata

:returns: A `.Variable` object
'''
    len_data = m.lhd[lblrec] - m.lhd[lbext]

    if len_data == 0:
        return None

    # Set the data type
    if m.lhd[lbuser1] == 1:
        datatype = m.file_float32
    elif m.lhd[lbuser1] == 2:
        datatype = m.file_int32
    elif m.lhd[lbuser1] == 3:
        # No idea if this works, or is valid code, or if I should even
        # consider it at all. Were logicals ever made part of PP? dch
        datatype = numpy.dtype('>b4')
    else:
        # Unknown data type => assume real
        datatype = m.file_float32

    # Create the data object
    parameters = {'_file'     : ppfile,
                  '_start'    : ppfile.tell(),
                  '_shape'    : (m.lhd[lbrow], m.lhd[lbnpt]),
                  '_dtype'    : datatype,
                  }    

    if m.bhd[bmks] != -1.0e30:
        parameters['_FillValue'] = m.bhd[bmdi]

    if abs(m.bhd[bdatum]) > ATOL():
        parameters['add_offset'] = m.bhd[bdatum]

    if not abs(m.bhd[bmks]-1.0) <= ATOL() + RTOL():
        parameters['scale_factor'] = m.bhd[bmks]

    data = Variable(**parameters)

    # Skip data and 4th block control word to get to the start of the
    # next PP field in the file
    ppfile.seek(len_data*_word_size, os.SEEK_CUR)

    # Return the data object
    return data
#--- End: def

def _coordinate_data(c, m=None,
                     data=None, bounds=None, climatology=False, 
                     coord_type=None, extra_type=None,
                     parameters={}):
    '''
    
Create a coordinate variable's data, either as a VariableCalc object
(which creates a regularly spaced numpy array from parameters) object
or a numpy array.

:parameter m: A PP field's complete metatdata
:type m: PPMetadata

:returns: The updated coordinate variable.
'''
    if c is None:
        return None

    create_bounds = True

    #-----------------------------------
    # Set time data from vtime and dtime
    #-----------------------------------
    if coord_type == 't':
        # This PP field's data array does not have time as one of
        # its two axes, so create a size 1 dimension coordinate for time
        data, bounds, climatology = None, None, False
        ib = m.lbtim_ib
        if ib <= 1 or m.vtime >= m.dtime:
            data = numpy.array([m.vtime])
        else:
            data = numpy.array([0.5*(m.vtime + m.dtime)])
            if ib <= 3:
                bounds = numpy.array([[m.vtime, m.dtime]])
                if ib == 3:
                    climatology = True
    #--- End: if

    # ------------------------------------------------
    # Set data from given values
    # ------------------------------------------------
    if data is not None:
        # Create a coordinate data object
#        c._data = Data(numpy.array(data))
        c._data = Data(data)

    if bounds is not None:
        # Create a coordinate bounds data object
        c.bounds       = CoordinateBounds()
        c.bounds._data = Data(bounds)
        if climatology:
            c.climatology = climatology

    # ------------------------------------------------
    #
    # ------------------------------------------------
    if coord_type == 'x':
        if abs(m.bhd[bdx]) <= ATOL() or m.timeseries: # DCH 
            # Create a coordinate data object from the field's extra
            # data
            try:
                c._data = Data(numpy.array(m.extra[coord_type]))
            except KeyError: 
                pass
            
            # Create a coordinate bounds data object from the field's
            # extra data
            try:
                m.extra[coord_type+'_bounds']
            except KeyError: 
                pass
            else:
                c.bounds       = CoordinateBounds()
                c.bounds._data = Data(numpy.array(m.extra[coord_type+'_bounds']))
                
        else:
            # Find, from the PP header, parameters for creating
            # coordinate data and coordinate bounds data objects
            parameters = {'_pseudo_origin': m.bhd[bzx],
                          '_delta'        : m.bhd[bdx],
                          '_shape'        : (m.lhd[lbnpt],),
                          '_dtype'        : m.float32,
                          }
            if m.ix in (13, 40, 99):
                create_bounds = False                    

            if (m.ix == 11 and
                parameters['_pseudo_origin'] + parameters['_delta']*parameters['_shape'][0] > 360.0):
                parameters['_pseudo_origin'] -= 360.0

    elif coord_type == 'y':
        if abs(m.bhd[bdy]) <= ATOL() or m.timeseries: # DCH Set a
            # Create a coordinate data object from the field's extra
            # data
            try:
                c._data = Data(numpy.array(m.extra[coord_type]))
            except KeyError: 
                pass
            
            # Create a coordinate bounds data object from the field's
            # extra data
            try:
                m.extra[coord_type+'_bounds']
            except KeyError: 
                pass
            else:
                c.bounds       = CoordinateBounds()
                c.bounds._data = Data(numpy.array(m.extra[coord_type+'_bounds']))
                
        else:
            # Find, from the PP header, parameters for creating
            # coordinate data and coordinate bounds data objects
            parameters = {'_pseudo_origin': m.bhd[bzy],
                          '_delta'        : m.bhd[bdy],
                          '_shape'        : (m.lhd[lbrow],),
                          '_dtype'        : m.float32,
                          }
            if m.iy in (13, 40, 99):
                create_bounds = False         

            if (m.iy == 11 and
                parameters['_pseudo_origin'] + parameters['_delta']*parameters['_shape'][0] > 360.0):
                parameters['_pseudo_origin'] -= 360.0
    #--- End: if
        
    if extra_type is not None:
        try:
            m.extra[extra_type] 
        except KeyError: 
            pass
        else:
            if not extra_type.endswith('bounds'):
                # Create a coordinate data object from the field's extra
                # data
                c._data = Data(numpy.array(m.extra[extra_type])) # dch
            else:
                # Create a coordinate bounds data object from the field's
                # extra data
                c.bounds       = CoordinateBounds()
                c.bounds._data = Data(numpy.array(m.extra[extra_type]))
    #--- End: if

    if parameters:
        # Create a coordinate data object
        c._data = Data(VariableCalc(**parameters))

        # Create a coordinate bounds data object
        if create_bounds:
            parameters['_shape'] += (2,)            
            c.bounds       = CoordinateBounds()
            c.bounds._data = Data(VariableCalcBounds(**parameters))
    #--- End: if 

    # Return coordinate
    return c
#--- End: def

def _create_Coordinate(f, key, axis_code=None, m=None, pubattr={},
                       data=None, bounds=None, climatology=False, 
                       coord_type=None, extra_type=None,
                       parameters=None, dimensions=[]):
    '''
    
Create a coordinate variable.

:parameter m: A PP field's complete metatdata
:type m: PPMetadata

'''
    c = Coordinate()

    if axis_code:
        # Set attributes
        c.axis =           _coord_axis[axis_code]
        c.positive =       _coord_positive[axis_code]
        c.standard_name =  _coord_standard_name[axis_code]
        c.units =          _coord_units[axis_code]
        
        if c.standard_name is None:
            c.long_name = _coord_long_name[axis_code]
            del c.standard_name

        # Special cases
        if axis_code in (20, 23):
            # Time
            c.units    = m.time_units
            c.calendar = m.calendar
        elif 100 < m.lhd[lbcode] < 10000:            
            # Rotated lat-lon
            if axis_code == 10:
                c.units         = 'degrees'
                c.standard_name = 'grid_latitude'
            elif axis_code == 11:
                c.units         = 'degrees'
                c.standard_name = 'grid_longitude'
        
        # Remove None attributes        
        [c.delpub(attr) for attr in ('axis', 'positive') 
         if c.getpub(attr) is None]
    #--- End: if

    # Apply extra public attributes, or overwrite existing ones
    for attr, value in pubattr.iteritems():
        if value is not None:
            c.setpub(attr, value)
        else:
            try:
                c.delpub(attr)
            except AttributeError: 
                pass
    #--- End: for

    # data
    c = _coordinate_data(c, m=m,
                         data        = data,
                         bounds      = bounds, 
                         climatology = climatology,
                         coord_type  = coord_type,
                         extra_type  = extra_type, 
                         parameters  = parameters)

    f.space.dimensions[key] = dimensions

    if key.startswith('d'): # d for dim
        # Find the dimension size
        try:
            dimension_size = c.size
        except AttributeError:
            if coord_type == 'x':
                dimension_size = m.lhd[lbnpt]
            elif coord_type == 'y':
                dimension_size = m.lhd[lbrow]
            else:
                raise ValueError,"Can't find dimension size for coord_type '%s'" % \
                    coord_type
        #--- End: try
        f.space.dimension_sizes[key] = dimension_size

        # Find the direction of the dimension coordinate 
        f.space.increasing[key] = dimension_direction(c)
    #--- End: if

    c = set_Data_attributes(f, key, c)

    f.space[key] = c

    # Set the coordinate's netCDF variable name
    c.ncvar = getattr(c, 'standard_name', getattr(c, 'long_name', None))
        
    return c
#--- End: def

def _decode_exp(lbexp):
    '''
    
Decode the integer value of LBEXP in the PP header into a 5-character
runid.

If this value has already been decoded, then it will be returned from
the cache, otherwise the value will be decoded and then added to the
cache.

:parameter lbexp: The value of LBEXP in the PP header
:type lbexp: int

:returns: A 5-character string derived from LBEXP. If LBEXP is a
          negative integer then that number is returned as a string.
'''
    try:
        # Try to return a cached decoding of this LBEXP
        return _cache['runid'][lbexp]
    except KeyError:
        # This LBEXP hasn't been decoded yet
        pass

    if lbexp < 0:
        return str(lbexp)

    # Convert LBEXP to a binary string, filled out to 30 bits with
    # zeros
    bits = bin(lbexp)
    bits = bits.lstrip('0b').zfill(30)

    # Step through 6 bits at a time, converting each 6 bit chunk into
    # a decimal integer, which is used as an index to the characters
    # lookup list.
    runid = []
    for i in xrange(0,30,6):
        index = int(bits[i:i+6], 2)
        try:
            runid.append(_characters[index])
        except IndexError:
            # Can't translate index >= 36
            pass
    #--- End: for
    runid = ''.join(runid)

    # Enter this runid into the cache
    _cache['runid'][lbexp] = runid

    # Return the runid
    return runid
#--- End: def 

def _set_axis_codes(m=None):
    '''
    
Find the axis codes for the dimensions occupying the T, Z, Y and X
dimension positions of the PP header.

:parameter m: A PP field's complete metatdata
:type m: PPMetadata

:returns: The PP metadata object updated with the axis codes set in
          its `it`, `iz`, `iy` and `ix` attributes.
'''
    code = m.lhd[lbcode]
    if code in (1, 101, 2, 102):
        m.ix, m.iy = 11, 10
    elif code >= 10000:
        temp,   ix = divmod(code, 10000)
        m.ix, m.iy = divmod(ix, 100)
    else:
        m.ix, m.iy = None, None

    # Determine if we have a site-time cross section and, if so, if it
    # is a timeseries
    m.site_time_cross_section = False
    m.timeseries              = False
    if (m.ix, m.iy) == (13, 23):
        m.site_time_cross_section = True
        m.timeseries = m.lhd[lbuser3] == m.lhd[lbrow]

    # Set iv from LBVC
    try:
        m.iv = _lbvc_codes[m.lhd[lbvc]]
    except KeyError:
        m.iv = None

    # Set it
    if m.calendar == 'gregorian':
        m.it = 20
    else:
        m.it = 23

    return m
#--- End def

def _pp_header(ppfile):
    '''

Read the header of the PP field at the current position in the given
PP file and create some derived quantities.

:parameter ppfile: An open file object for a PP file.
:type ppfile: file

:returns: A :class:`~cf.pp.PPMetadata` object containing all of the metadata for the
          current PP field in the input file, including some derived
          metadata quantities, such as 'calendar', 'time_units',
          'timeseries', etc.
'''

    m = PPMetadata()

    # Read 1st block control word
    bcw1 = numpy.fromfile(ppfile, dtype=m.file_int32, count=1)
    
    # Stop now if we have reached the end of the file
    if not bcw1:
        return None

    # Change data types if we have a wrong endian PP file. ('Wrong'
    # simply means different to that assumed by class PPMetadata.)
    if bcw1[0] == 65536:        
        m.file_float32 = m.float32.newbyteorder('S')
        m.file_int32   = m.int32.newbyteorder('S')
        m.file_str32   = m.str32.newbyteorder('S')
    elif bcw1[0] != 256:
        raise RuntimeError,"Not a PP field: First block control word = %s" % bcw1[0]

    # Read the integer and real headers
    m.lhd = numpy.fromfile(ppfile, dtype=m.file_int32  , count=45)
    m.bhd = numpy.fromfile(ppfile, dtype=m.file_float32, count=21)

    # Set some derived metadata quantities. Note that we need to this
    # before we get any extra data.
    m.lbtim_ia, ib         = divmod(m.lhd[lbtim], 100)
    m.lbtim_ib, m.lbtim_ic = divmod(ib, 10)

    # Set units to 'days since 1860-12-1' for now, since field
    # aggregation can't deal with equivalent but non-equal units, yet.
    m.time_units = 'days since 1860-12-1'
    if m.lbtim_ic == 1:
        m.calendar = 'gregorian'
    else:
        m.calendar = '360_day'

    # For speed, try to return a cached numeric time value, otherwise
    # calculate one.
    v = (m.lhd[lbyr], m.lhd[lbmon], m.lhd[lbdat], m.lhd[lbhr], m.lhd[lbmin])
    d = (m.lhd[lbyrd], m.lhd[lbmond], m.lhd[lbdatd], m.lhd[lbhrd], m.lhd[lbmind])
    
    vdatetime = datetime(v[0], v[1], v[2], v[3], v[4])
    ddatetime = datetime(d[0], d[1], d[2], d[3], d[4])

    for attr, t, mm in zip(('vtime', 'dtime'),
                           ( vdatetime ,  ddatetime),
                           ( v, d)):
        try:
            value = _cache['time'][mm]
        except KeyError:
            value = netCDF4.date2num(t, m.time_units, m.calendar)
            _cache['time'][mm] = value
        setattr(m, attr, value)
    #--- End: for

    version, model = divmod(m.lhd[lbsrce], 10000)
    if model > 0:
        m.model = 'UM'
    if version > 0:
        m.version = str(version)
    
    m = _set_axis_codes(m=m)

    # Return the metadata
    return m
#--- End: def

def _test_pp_condition(pp_condition, m=None):
    '''

Test whether or not a field satisfies the condition specified for a PP
STASH code to standard name conversion.

:keyword m: A PP field's complete metatdata
:type m: PPMetadata

:returns: True if a field satisfies the condition specified, False
          otherwise.
'''
    if pp_condition is '':
        # Return True if no condition is set
        return True

    # Still here?
    if pp_condition == 'true_latitude_longitude':
        if m.lhd[lbcode] in (1, 2):
            return True
        # Check pole location in case of incorrect LBCODE
        if (abs(m.bhd[bplat]-90.0) <= ATOL() + RTOL()*90.0 and 
            abs(m.bhd[bplon]) <= ATOL()):
            return True

    elif pp_condition == 'rotated_latitude_longitude':
        if m.lhd[lbcode] in (101, 102):
            return True
        # Check pole location in case of incorrect LBCODE
        if not (abs(m.bhd[bplat]-90.0) <= ATOL() + RTOL()*90.0 and 
                abs(m.bhd[bplon]) <= ATOL()):
            return True
        
    else:
        raise ValueError,"Unknown PP condition in for STASH code %d: '%s'" % \
            (m.lhd[lbuser4], pp_condition)

    # Still here? Then the condition has not been satisfied.
    return False
#--- End: def

def _test_umversion(valid_from, valid_to, m=None):
    '''

Test whether or not the UM version applicable to this field is within
the given range.

If possible, the UM version is derived from the PP header and stored
in the metadata object. otherwise it is taken from the UMVERSION
environment variable.

:returns: True the UM version applicable to this field is within the
          range, False otherwise.
'''
    try:
        umversion = m.version
    except AttributeError:
        try:            
            umversion = _UMVERSION
        except NameError:
            raise RuntimeError,"UM version is not set in PP header nor the environment"
    #--- End: try

    if valid_from is '':
        valid_from = None

    if valid_to is '':
        if valid_from <= umversion:
            return True 
    elif valid_from <= umversion <= valid_to:
        return True 

    return False
#--- End: def

def _field_attributes(f, m=None):
    '''

Set public and private attributes on the space (including the standard
name).
'''
    f.setpriv('pp'  , m)
    f.setpriv('file', m.file)

    f._FillValue = m.bhd[bmdi]
    f.setpub('Conventions', __Conventions__)    
    f.setpub('history'    , m.history)
    f.setpub('runid'      , _decode_exp(m.lhd[lbexp]))

#    if abs(m.bhd[bdatum]) > ATOL():
#        f.setpub('add_offset', m.bhd[bdatum])
#
#    if not abs(m.bhd[bmks]-1.0) <= ATOL() + RTOL():
#        f.setpub('scale_factor', m.bhd[bmks]) # dch: check how this interacts with f.units       

    try:
        f.setpub('source', m.model)
    except AttributeError:
        pass
    else:
        try:
            f.setpub('source', '%s vn%s' % (f.getpub('source'), m.version))
        except AttributeError:
            pass        

    try:
        f.setpub('title', m.extra['title'])
    except KeyError:
        pass

    # ----------------------------------------------------------------
    # Set the field's standard_name, long_name and units from its
    # STASH code and possibly other properties
    # ----------------------------------------------------------------
    stash    = m.lhd[lbuser4]
    submodel = m.lhd[lbuser7]

    f.setpub('stash_code', stash)
    f.setpub('submodel'  , submodel)

    m.cf_info = {}

    # The STASH code has been set in the PP header, so try to find
    # its standard_name from the conversion table
    try:
        lines = _stash2standard_name[(submodel, stash)]
    except KeyError:
        if stash:
            f.long_name = 'PP_%d_%d' % (submodel, stash)
        else:
            f.long_name = 'PP_%d_fc%d' % (submodel, m.lhd[lbfc])
    else:
        for (long_name, 
             units,
             valid_from,
             valid_to, 
             standard_name,
             cf_info,
             pp_condition) in lines:

            # Check that conditions are met             
            version_ok      = _test_umversion(valid_from, valid_to, m=m)
            pp_condition_ok = _test_pp_condition(pp_condition, m=m)
            if not (version_ok and pp_condition_ok):                    
                continue

            # Still here? Then we have our standard_name, etc.
            f.long_name = long_name.rstrip()
            f.units     = units
            if standard_name:
                f.standard_name = standard_name

            m.cf_info = cf_info
                
            break
        #--- End: for            
    #--- End: try

    # If there is no standard name then set a private aggregating name
    # based on the submodel and STASHcode (and maybe the field code)
    try:
        f.standard_name
    except AttributeError:
        if stash:
            f.setpriv('aggregating_name', 'PP_%d_%d' % (submodel, stash))
        else:
            f.setpriv('aggregating_name', 'PP_%d_fc%d' % (submodel, m.lhd[lbfc]))
    #--- End: try

    # If there is no long name then set a default
    try:
        f.long_name
    except AttributeError:
        if stash:
            f.long_name = 'PP_%d_%d' % (submodel, stash)
        else:
            f.long_name = 'PP_%d_fc%d' % (submodel, m.lhd[lbfc])

    # Set the field's coordinate's netCDF variable name
    f.ncvar = getattr(f, 'standard_name', f.getpriv('aggregating_name', None))
        
#
#        try:
#            f.getpub('long_name')
#            try:   # dch this is a nightmare
#                if not (f.getpub('units') == '%' and f.getpub('scale_factor') == 0.01):
#                    f.setpub('units', '%s %s' % (f.getpub('scale_factor'), f.getpub('units')))
#                    f.delpub('scale_factor')
#            except AttributeError:
#                pass
#        except AttributeError:
#            # Default long name in the absence of a conversion
#            f.setpub('long_name', 'STASHcode=%s' % stash)
#    #--- End: if

    return f
#--- End: def

def read_stash2standard_name(table=None, delimiter='!'):
    ''' 

Read the STASH to standard_name conversion table, as found at
http://puma.nerc.ac.uk/STASH_to_CF/STASH_to_CF.txt, into a dictionary.

If `table` is None then table will be looked for at
`os.path.join(os.path.dirname(cf.__file__), 'etc/STASH_to_CF.txt')`

:returns: A dictionary of tuples of lists

*Examples*

>>> stash2sn = read_stash2standard_name()
>>> print stash2sn[(1,2)]
(['U COMPNT OF WIND AFTER TIMESTEP     ',
  'm s-1',
  '',
  '',
  'eastward_wind',
  '',
  'true_latitude_longitude'],
 ['U COMPNT OF WIND AFTER TIMESTEP     ',
  'm s-1',
  '',
  '',
  'x_wind',
  '',
  'rotated_latitude_longitude'])
'''

    # 0  Model           
    # 1  STASH code      
    # 2  STASH name      
    # 3  units           
    # 4  valid from UM vn
    # 5  valid to   UM vn
    # 6  standard_name   
    # 7  CF extra info   
    # 8  PP extra info

    if table is None:
        # Use default conversion table
        package_path = os.path.dirname(__file__)
        table = os.path.join(package_path, 'etc/STASH_to_CF.txt')

    lines = csv.reader(open(table, 'r'), 
                       delimiter=delimiter, skipinitialspace=True)

    raw_list = []
    [raw_list.append(line) for line in lines]

    # Get rid of comments
    for line in raw_list[:]:
        if line[0].startswith('#'):
            raw_list.pop(0)
            continue
        break
    #--- End: for

    # Convert to a dictionary which is keyed by (submodel, STASHcode)
    # tuples

    (model, stash, name,
     units, 
     valid_from, valid_to,
     standard_name, cf, pp_info) = range(9)
        
    stash2sn = {}
    for x in raw_list:
        key = (int(x[model]), int(x[stash]))
        try:            
            cf_info = {}
            if x[cf]:
                for d in x[7].split():
                    if d.startswith('height='): 
                        cf_info['height'] = re.split(_number_regex, d,
                                                     re.IGNORECASE)[1:4:2]
                        if cf_info['height'] == '':
                            cf_info['height'][1] = '1'

                    if d.startswith('below_'):
                        cf_info['below'] = re.split(_number_regex, d,
                                                     re.IGNORECASE)[1:4:2]
                        if cf_info['below'] == '':
                            cf_info['below'][1] = '1'

                    if d.startswith('where_'):         
                        cf_info['where'] = d.replace('where_', 'where ', 1)
                    if d.startswith('over_'):         
                        cf_info['over'] = d.replace('over_', 'over ', 1)

            x[cf] = cf_info                    
        except IndexError:
            pass

        x[pp_info] = x[pp_info].rstrip()

        line = (x[name:],)

        try:
            stash2sn[key] += line
        except KeyError:
            stash2sn[key] = line

    #--- End: for

    return stash2sn
#--- End: def

# Create the STASH code to standard_name conversion dictionary at import time
_stash2standard_name = read_stash2standard_name()


def read(infile, close=False, exact=False, extract={}):
    ''' 

Read spaces from an input PP file on disk.

The file may be big or little endian.

All data arrays from all space components are not read from disk but
are stored in the returned spaces as file pointers.


:argument filename: An open file object, or a string giving the file
                    name, from which to read spaces.

:type filename: file or str

   
:parameter exact: *Optional*. If True then interpret the keywords from
                  `**kwargs` as exact phenomenon names. Refer to
                  `cf.space.match` for details.

:type exact: bool


:parameter **kwargs: *Optional*. Other keyword arguments defining
                     space selection based on phenomena criteria. If
                     none are set then all spaces are
                     selected. Keywords and their arguments are set as
                     for `cf.space.match`.


:returns: A list of spaces (a `cf.FieldList` object).

**Examples**

>>> s = cf.pp.read('file.pp')

'''

    # coord_type: A short string which acts as .... Valid values are
    # 't'  : time 
    #
    # 'v'  : an axis other than time which has the the same value for
    #        all data points, as described by LBVC
    #
    # 'p'  : pseudolevel
    #
    # 'rwl': radiation wavelength
    #
    # 'x'  : The X axis of the data array
    #
    # 'y'  : The Y axis of the data array

    # ------------------------------------------------------------
    # Preliminaries
    # ------------------------------------------------------------
    # Open the file, if it is not already open
    try:
        ppfile = open(infile, 'rb') 
    except TypeError:
        ppfile = infile

#    rr=open('../etc/arse','r')

    # Initialize the output list of fields
    selected_fields_in_file = FieldList()

    history = '%s Converted from PP by cf-python v%s' % \
        (time.strftime('%Y-%m-%d %H:%M:%S UTC', time.gmtime()),
         __version__)
    
    # ****************************************************************
    # Loop round each PP field in the file, reading it into a field
    # ****************************************************************
    while True:

        dimN   = 0  # Counter for dimension coordinates
        auxN   = 0  # Counter for auxiliary coordinates
        transN = 0  # Counter for transforms

        # ============================================================
        # Get the next PP field's metadata and data and find some
        # derived metatdata quantities, such as 'calendar',
        # 'time_units', 'timeseries', etc.
        # ============================================================
        # Get the header (lhd and bhd)
        m = _pp_header(ppfile)

        # Stop now if we have reached the end of the file
        if m is None:
            break

        # Get the data
        field_data_array = _field_data(ppfile, m=m)

        # Get any extra data
        m = _extra_data(ppfile, m=m)

        m.history = history
        m.file    = ppfile.name

        # ============================================================
        # Initialize the field. The involves setting the field's
        # attributes and data and giving it an empty space, for now.
        # ============================================================
        f       = Field()
        f.space = Space()
        f.space.increasing = {}

        # Set the field's attributes
        f = _field_attributes(f, m=m)

        axis_name = {}
        axis_dim = {}
        
        # ============================================================
        # Create the field's space
        # ============================================================
       
        # ------------------------------------------------------------
        # Create a TIME dimension coordinate
        # ------------------------------------------------------------
        dim  = 'dim%(dimN)d' % locals()
        dimN += 1         # Increment dimension number
        tdim = dim      

        # Set coordinate type and axis code
        if m.ix in (20, 23):
            coord_type = 'x'
            axis_code  = m.ix
            # Flag that we have already created an 'X' coordinate
            m.ix = None
        elif m.iy in (20, 23):
            coord_type = 'y'
            axis_code  = m.iy
            # Flag that we have already created a 'Y' coordinate
            m.iy = None
        else:
            coord_type = 't'
            axis_code  = m.it
            
        # Set the coordinate's data
        if m.timeseries:
            # This PP field is a timeseries. The validity time is
            # taken to be the time for the first sample, the data time
            # for the last sample, with the others evenly between.
            p = {'_shape' : (m.lhd[lbrow],),
                 '_dtype' : m.float32,}
            p['_delta'] = (m.dtime-m.vtime)/(p['_shape'][0]-1)
            p['_pseudo_origin'] = m.vtime - p['_delta']
        
            coord = _create_Coordinate(f, tdim, axis_code, m=m, 
                                       parameters = p, 
                                       dimensions = [tdim])       
        else:
            coord = _create_Coordinate(f, tdim, axis_code, m=m,
                                       coord_type = coord_type, 
                                       dimensions = [tdim])
       

        # Save the dimension name
        if coord_type == 'x':
            data_dim0 = tdim
        elif coord_type == 'y' or m.timeseries:
            data_dim1 = tdim

        # Set the axis name for cell methods
        axis_name['t'] = coord.standard_name
        axis_dim['t']  = tdim
        
        # ------------------------------------------------------------
        # Create a 'VERTICAL' dimension coordinate. This size 1
        # coordinate will often by a vertical coordinate such as
        # pressure, but could be anything, such as longitude.
        # ------------------------------------------------------------
        if m.iv is not None: 
            dim  = 'dim%(dimN)d' % locals()
            dimN += 1         # Increment dimension number
            vdim = dim      

            try:
                height, units = m.cf_info['height']
            except (KeyError, TypeError):
                # Create the height coordinate from the PP header
                if m.lhd[lbvc] == 9:
                    # atmosphere_hybrid_sigma_pressure_coordinate
                    pstar  = 1.0e5
                    data   = m.bhd[blev] + m.bhd[bhlev]/pstar
                    bounds = numpy.array([[m.bhd[brlev]  + m.bhd[bhrlev]/pstar, 
                                           m.bhd[brsvd1] + m.bhd[brsvd2]/pstar]])
                elif m.lhd[lbvc] == 65:
                    # atmosphere_hybrid_height_coordinate 
                    pass
                else:
                    data   = m.bhd[blev]
                    bounds = numpy.array([[m.bhd[brlev], m.bhd[brsvd1]]])
                    if equals(bounds[0,0], bounds[0,1]):
                        bounds = None

                coord = _create_Coordinate(f, vdim, axis_code=m.iv, m=m,
                                           data       = numpy.array([data]),
                                           bounds     = bounds,
                                           dimensions = [vdim])    
            else:               
                # Create the height coordinate from the information
                # given in the STASH to standard_name conversion table
                coord = _create_Coordinate(f, vdim, axis_code=15, m=m,
                                           pubattr    = {'units': units},
                                           data       = numpy.array([height], 
                                                                    dtype=float),
                                           dimensions = [vdim])
            #--- End: try

            # Set the axis name for cell methods
            axis_name['v'] = coord.standard_name
            axis_dim['v']  = vdim
        #--- End: if

        # ------------------------------------------------------------
        # Create a PSEUDOLEVEL dimension coordinate
        # ------------------------------------------------------------
        if m.lhd[lbuser5] > 0: 

            dim  = 'dim%(dimN)d' % locals()
            dimN += 1         # Increment dimension number
            pdim = dim      
            
            coord = _create_Coordinate(f, pdim, axis_code=40, m=m, 
                                       data       = numpy.array([m.lhd[lbuser5]]),
                                       dimensions = [pdim])
        
            # Set the axis name for cell methods (as yet there is no
            # standard_name for this, so use the long_name)
            axis_name['p'] = coord.standard_name #long_name
            axis_dim['p']  = pdim
        #--- End: if

        # ------------------------------------------------------------
        # Create a RADIATION WAVELENGTH dimension coordinate
        # ------------------------------------------------------------
            #dch : do this instead of VERTICAL, I wonder
        try:
            rwl, units = m.cf_info['below']
        except (KeyError, TypeError):
            pass
        else:
            
            dim   = 'dim%(dimN)d' % locals()
            dimN += 1         # Increment dimension number
            rwldim = dim      
            
            # Create the radiation wavelength coordinate
            pubattr = {'standard_name' : 'radiation_wavelength', 
                       'units'         : units,
                       }
            coord = _create_Coordinate(f, rwldim, axis_code=None, m=m,
                                       pubattr    = pubattr,
                                       data       = numpy.array([rwl], 
                                                                dtype=float),
                                       bounds     = numpy.array([[0.0, rwl]]),
                                       dimensions = [rwldim])
            
            # Set the axis name for cell methods
            axis_name['rwl'] = coord.standard_name
        #--- End: try

        # ------------------------------------------------------------
        # Create X and Y dimension coordinates, if they haven't
        # already been created. These coordinates represent the 'x'
        # and 'y' dimensions of the data array. In the case of cross
        # sections, one of them may have already been created.
        # ------------------------------------------------------------
        # Set coordinate type and axis code
        for coord_type, axis_code in zip(('y' , 'x' ),
                                         (m.iy, m.ix),
                                         ):
            # Skip this coordinate if it has already been created
            if axis_code is None:
                continue    

            dim   = 'dim%(dimN)d' % locals()           
            dimN += 1         # Increment dimension number       

            coord = _create_Coordinate(f, dim, axis_code, m=m,
                                       coord_type=coord_type,
                                       dimensions=[dim])
        
            if coord_type == 'y':
                ydim      = dim
                data_dim0 = ydim
            else:
                xdim      = dim
                data_dim1 = xdim
 
            # Set the axis name for cell methods
            axis_name[coord_type] = coord.standard_name
        #--- End: for

        # ============================================================
        # Create 1-d, size 1 auxiliary coordinates
        # ============================================================

        # ------------------------------------------------------------
        # Atmosphere hybrid sigma pressure coordinate components
        # ------------------------------------------------------------        
        if m.lhd[lbvc] == 9:
            for long_name, value, units, bounds in zip(
                ('atmosphere_hybrid_sigma_pressure_coordinate_ak',      
                 'atmosphere_hybrid_sigma_pressure_coordinate_bk'),
                (m.bhd[bhlev],                     # value
                 m.bhd[blev]),
                ('Pa',                             # units
                 '1'),
                ([m.bhd[bhrlev], m.bhd[brsvd2]],   # bounds
                 [m.bhd[brlev] , m.bhd[brsvd1]]),
                ):
                aux   = 'aux%(auxN)d' % locals()           
                auxN += 1         # Increment auxiliary number
                
                pubattr = {'standard_name': long_name,
                           'units'        : units,
                           }                
                coord = _create_Coordinate(f, aux, axis_code=None, m=m,
                                           pubattr    = pubattr, 
                                           data       = numpy.array([value]),
                                           bounds     = numpy.array([bounds]),
                                           dimensions = [vdim])
            #--- End: for
        #--- End: if

        # ------------------------------------------------------------
        # Create a forecast_reference_time auxiliary coordinate
        # ------------------------------------------------------------
        if m.lbtim_ib == 1: 
            aux   = 'aux%(auxN)d' % locals()           
            auxN += 1         # Increment auxiliary number
       
            pubattr = {'standard_name': 'forecast_reference_time',
                       'units'        : m.time_units,
                       'calendar'     : m.calendar,
                      }
            coord = _create_Coordinate(f, aux, axis_code=None, m=m,
                                       pubattr    = pubattr, 
                                       data       = numpy.array([m.dtime], 
                                                                dtype=float),
                                       dimensions = [tdim])           
        #--- End: if

        # -------------------------------------------------------
        # Create a model_level_number auxiliary coordinate
        # -------------------------------------------------------
        lev = m.lhd[lblev]
        if (m.iv is not None and 
            m.lhd[lbvc] in (2, 9, 65) and 
            lev not in (7777, 8888, 9999)): # dch check 7777
            aux   = 'aux%(auxN)d' % locals()           
            auxN += 1         # Increment auxiliary number

            coord = _create_Coordinate(f, aux, axis_code=5, m=m, 
                                       pubattr    = {'axis': None}, 
                                       data       = numpy.array([lev]), 
                                       dimensions = [vdim])
        #--- End: if

        # ------------------------------------------------------------
        # Create domain title auxiliary coordinate
        # ------------------------------------------------------------
        try:            
            domain_titles = m.extra['domain_title']
        except KeyError:
            pass
        else:
            dim = None
            if ix == 13:
                dim = xdim
            elif iy == 13:
                dim = ydim
            if dim:
                aux   = 'aux%(auxN)d' % locals()           
                auxN += 1                        # Increment auxiliary number

                pubattr = {'units'    : 'string',
                           'long_name': 'domain title',
                           }
                coord = _create_Coordinate(f, aux, axis_code=None, m=m, 
                                           data       = numpy.array(domain_titles), # dch array, or not?
                                           pubattr    = pubattr,
                                           dimensions = [dim])
        #--- End: try

        # ------------------------------------------------------------
        # Create 2-d unrotated latitude and longitude auxiliary
        # coordinates.
        # ------------------------------------------------------------
        if f.space[ydim].units == f.space[xdim].units == 'degrees':

            if f.space[ydim].standard_name == 'grid_longitude':
                rotated_lon = f.space[ydim]
                rotated_lat = f.space[xdim]
            else:
                rotated_lon = f.space[xdim]
                rotated_lat = f.space[ydim]

            # Try to find the appropriate unrotated latitude and
            # longitude arrays in the cache, to avoid having to create
            # them.
            found_cached_latlon = False            
            for rot_lat, rot_lon in _cache['latlon']:
                if (rot_lon._data.equals(rotated_lon._data) and 
                    rot_lat._data.equals(rotated_lat._data)):
                    
                    found_cached_latlon = True

#                    latlon = _cache['latlon'][(rot_lat, rot_lon)] 
                    lat, lon = _cache['latlon'][(rot_lat, rot_lon)] 

                    if f.space[ydim].standard_name == 'grid_longitude':
#                         latlon = numpy.transpose(latlon, axes=(0,2,1))
                         lat = numpy.transpose(lat, axes=(0,2,1))
                         lon = numpy.transpose(lon, axes=(0,2,1))

                    break
            #--- End: for

            # Create the unrotated latitude and longitude arrays if we
            # couldn't find them in the cache
            if not found_cached_latlon:
#                print  'CALC'
                lat, lon = unrotated_latlon(rotated_lat.varray, rotated_lon.varray,
                                            m.bhd[bplat], m.bhd[bplon]) 

                # Add the newly created unrotated latitude and
                # longitude auxiliary coordinates to the cache
                _cache['latlon'][(rotated_lat, rotated_lon)] = (lat, lon)

                if f.space[ydim].standard_name == 'grid_longitude':
#                    latlon = numpy.transpose(latlon, axes=(0,2,1))
                    lat = numpy.transpose(lat, axes=(0,2,1))
                    lon = numpy.transpose(lon, axes=(0,2,1))
            #--- End: if
             
            # Create the tranform
            transform_id = 'trans%(transN)d' % locals()           
            transN += 1     # Increment transform number

            grid_mapping_name = 'rotated_latitude_longitude'

            transform = Transform()
            transform.name = grid_mapping_name
            transform['grid_mapping_name']         = grid_mapping_name
            transform['grid_north_pole_latitude']  = m.bhd[bplat]
            transform['grid_north_pole_longitude'] = m.bhd[bplon]
            try:
                f.space.transform
            except AttributeError:
                f.space.transform = CfDict()     
            f.space.transform[transform_id] = transform

            # Add the transform pointer to the dimension coordinates
            f.space[ydim].transform = transform_id
            f.space[xdim].transform = transform_id

            # Create the unrotated latitude and longitude coordinates        
            for (i, axis_code, standard_name, data,
                 units) in zip((  0            ,   1           ),
                               ( 10            ,  11           ),
                               ('latitude'     , 'longitude'   ),
                               ( lat           ,  lon          ),
                               ('degrees_north', 'degrees_east'),
                               ):
                aux   = 'aux%(auxN)d' % locals()           
                auxN += 1                        # Increment auxiliary number

                pubattr = {'standard_name': standard_name,
                           'units'        : units,
                           'axis'         : None,
                          }                
                coord = _create_Coordinate(f, aux, axis_code, m=m,
                                           pubattr    = pubattr,
                                           data       = data,
                                           dimensions = [ydim, xdim]) 

                # Add the transform pointer to the auxiliary
                # coordinates
                coord.transform = transform_id
             #--- End: for

        #--- End: if

        # ------------------------------------------------------------
        # Create auxiliary coordinates for the latitudes and
        # longitudes of the sites in site-time cross sections. (Don't
        # consider 'z' extra data because we don't know what it refers
        # to.)
        # ------------------------------------------------------------
        if m.site_time_cross_section:

            for axis_code, extra_type in zip((11 , 10 ),
                                             ('x', 'y')):
                try:
                    coord_type = '%s_domain_bounds' % extra_type
                    m.extra[coord_type]
                except KeyError:
                    try:
                        coord_type = '%s_domain_lower_bound' % extra_type
                        m.extra[coord_type]
                    except KeyError:
                        pass
                    else:           
                        # Create, from extra data, an auxiliary
                        # coordinate with data but no bounds, if the
                        # data noes not contain any missing values
                        file_position = ppfile.tell()
                        data = m.extra[coord_type][...]
                        # Reset the file pointer after reading the
                        # extra data into a numpy array
                        ppfile.seek(file_position, os.SEEK_SET)
                        if not numpy.any(data == _pp_rmdi): # dch also test in bmdi?   
                            aux   = 'aux%(auxN)d' % locals()           
                            auxN += 1                        # Increment auxiliary number
                            coord = _create_Coordinate(f, aux, axis_code, m=m, 
                                                       data       = numpy.array(data), # dch
                                                       pubattr    = {'axis': None}, 
                                                       dimensions = [xdim])# DCH      
                else:                                                                         # xdim?    
                    # Create, from extra data, an auxiliary coordinate                        # should   
                    # with 1) data and bounds, if the upper and lower                         # be       
                    # bounds have no missing values; or 2) data but no                        # the      
                    # bounds, if the upper bound has missing values                           # axis     
                    # but the lower bound does not.                                           # which    
                    file_position = ppfile.tell()                                             # has      
                    bounds = m.extra[coord_type][...]                                         # axis_code
                    # Reset the file pointer after reading the extra                          # 13       
                    # data into a numpy array
                    ppfile.seek(file_position, os.SEEK_SET)
                    data = None
                    if numpy.any(bounds[...,1] == _pp_rmdi): # dch also test in bmdi?
                        if not numpy.any(bounds[...,0] == _pp_rmdi): # dch also test in bmdi?
                            data = bounds[...,0]
                        bounds = None
                    else:
                        data = numpy.mean(bounds, axis=1)
                    if (data, bounds) != (None, None):
                        aux   = 'aux%(auxN)d' % locals()           
                        auxN += 1                        # Increment auxiliary number
                        
                        coord = _create_Coordinate(f, aux, axis_code, m=m,
                                                   data       = data,
                                                   bounds     = bounds, 
                                                   pubattr    = {'axis': None},
                                                   dimensions = [xdim]) # DCH      
                                                                        # xdim?    
                                                                        # should   
                                                                        # be       
                                                                        # the      
                                                                        # axis     
                                                                        # which    
                                                                        # has      
                                                                        # axis_code
                                                                        # 13       
               #--- End: try
           #--- End: for
       #--- End: if
        
        # ------------------------------------------------------------
        # Set the field's data dimension names
        # ------------------------------------------------------------
        f.space.dimensions['data'] = [data_dim0, data_dim1]

        # ------------------------------------------------------------
        # Set the field's data its attributes
        # ------------------------------------------------------------
        if field_data_array is not None:

            # Add a leading size 1 dimension for each dimension which
            # does not span the 2-d field_data_array. Some of these
            # may get removed later when we squash the field. The
            # order in which these are looped through is arbitrary,
            # but the order given ensures that the order is as
            # expected in the field.
            for coord_type in ('p', 'v', 't'):
                try:
                    dim = axis_dim[coord_type]
                except KeyError:
                    pass
                else:
                    field_data_array._shape = (1,) + field_data_array._shape
                    f.space.dimensions['data'].insert(0, dim)
            #--- End: for

            f._data = Data(field_data_array)          
            f  = set_Data_attributes(f, 'data', f)
        #--- End: if

        # ============================================================
        # Create the field's cell methods
        # ============================================================
        cell_methods = []
        f.space.nc_dimensions = {}

        proc       = m.lhd[lbproc]
        tmean_proc = 0
        if m.lbtim_ib in (2, 3):
            if proc-128 in (0, 64, 2048, 4096, 8192):
                tmean_proc = 128
                proc -= 128
                
        # ------------------------------------------------------------
        # Area cell methods
        # ------------------------------------------------------------
        if m.ix in (10, 11, 12) and m.iy in (10, 11, 12):
            if proc == 64 or m.cf_info.has_key('where'):
                cell_methods.append('area: mean')

            try:
                cell_methods.append(m.cf_info['where'])
            except (KeyError, TypeError):
                pass
            else:
                try:
                    cell_methods.append(m.cf_info['over'])
                except (KeyError, TypeError):
                    pass
#            for info in m.cf_info:                
#                if info.startswith('area: mean where_'):                    
#                    cell_methods.append(info.replace('where_', 'where ', 1))
#                elif info.startswith('over_'):
#                    cell_methods.append(info.replace('over_', 'over ', 1))
        # dch : do speial zonal mean as as in pp_cfwrite
        
        # ------------------------------------------------------------
        # Vertical cell methods
        # ------------------------------------------------------------
        if proc == 2048:
            v_axisname = axis_name['v']
            cell_methods.append('%s: mean' % v_axisname)
            f.space.nc_dimensions[vdim] = v_axisname

        # ------------------------------------------------------------
        # Time cell methods
        # ------------------------------------------------------------
        t_axisname = axis_name['t']
        f.space.nc_dimensions[tdim] = axis_name['t']
        if m.lbtim_ib in (0, 1):
            cell_methods.append('%s: point' % t_axisname)
        elif proc == 4096:
            cell_methods.append('%s: minimum' % t_axisname)
        elif proc == 8192:
            cell_methods.append('%s: maximum' % t_axisname)
        if tmean_proc == 128:
            if m.lbtim_ib == 2:
                cell_methods.append('%s: mean' % t_axisname)
            elif m.lbtim_ib == 3:
                cell_methods.append('%s: mean within years' % t_axisname)
                cell_methods.append('%s: mean over years' % t_axisname) 

#        print m.lhd[lbproc], proc, tmean_proc, cell_methods, m.lbtim_ib
        # ------------------------------------------------------------
        # Add the cell methods to the field
        # ------------------------------------------------------------
        if cell_methods:
            f.cell_methods = CellMethods()
            f.cell_methods = f.cell_methods.parse(' '.join(cell_methods), 
                                                  field=f)

        f.setpub('lbproc', m.lhd[lbproc])
            
        del f.space.increasing
   
        # ============================================================
        # The field is complete(!). Add it to the list of fields to be
        # returned if it matches any given phenomena criteria
        # ============================================================
        if f._match(exact, **extract):
            f.squash()
            selected_fields_in_file.append(f)

        # ============================================================
        # Now read the next PP field in the file
        # ============================================================
    #--- End: while

    # Close the PP file
    if close:
        ppfile.close()

    # Return the selected list of fields
    return selected_fields_in_file
#--- End: def
